/*
	pppcharge.c - program to start and stop charging of internet
	stuff
	Copyright 1999 Jonathan McDowell for Project Purple

	12/04/1999 - Started.
*/

#define VERSION "0.0.3"

#include <stdio.h>
#include <time.h>
#include <string.h>
#include <unistd.h>
#include <getopt.h>
#include <stdlib.h>

#include "parse.h"
#include "readcfg.h"

#ifdef DEBUG
#define CHARGINGFILE "pppcharge.log"
#define TMPCHARGINGFILE "pppcharge.log.tmp"
#define CFGFILE "pppctl.cfg"
#else
#define CHARGINGFILE "/var/log/pppcharge.log"
#define TMPCHARGINGFILE "/var/log/pppcharge.log.tmp"
#define CFGFILE "/etc/pppctl.cfg"
#endif

struct charginginf {
	char name[9];
	unsigned char action; /* 0 = dialup, 1 = hangup */
	time_t time;
};

struct chargeband {
	int startday, endday;
	int starthour, endhour;
	int startmin, endmin;
	float cost;
	struct chargeband *next;
};

/* int charge(time_t starttime, time_t endtime)*/

static int html=0, verbose=0, mincharge=0;
struct chargeband *chargelist=NULL;

void parserange(char *str, int *start, int *end)
{
	char *snd;

	if (*str=='*') {
		*start=*end=-1;
	} else if ((snd=strchr(str, '-'))==NULL) {
		*start=*end=atoi(str);
	} else {
		*snd=0;
		snd++;
		*start=atoi(str);
		*end=atoi(snd);
	}
}

struct chargeband *addchargeband(struct chargeband *current, char *str)
{
	char *day, *hour, *min, *cost;
	struct chargeband *new, *tmp;

	day=str;
	if ((hour=strchr(day, ' '))==NULL) {
		printf("Invalid chargeband.\n");
		return current;
	}
	*hour=0;
	hour++;
	if ((min=strchr(hour, ' '))==NULL) {
		printf("Invalid chargeband.\n");
		return current;
	}
	*min=0;
	min++;
	if ((cost=strchr(min, ' '))==NULL) {
		printf("Invalid chargeband.\n");
		return current;
	}
	*cost=0;
	cost++;

	if ((new=malloc(sizeof(struct chargeband)))==NULL) {
		printf("Couldn't get memory for new chargeband.\n");
		return current;
	} else {
		parserange(day, &(new->startday), &(new->endday));
		parserange(hour, &(new->starthour), &(new->endhour));
		parserange(min, &(new->startmin), &(new->endmin));
		new->cost=atof(cost);
		new->next=NULL;
	}

	if (current==NULL) {
		return new;
	} else {
		tmp=current;
		while (tmp->next!=NULL) tmp=tmp->next;
		tmp->next=new;
	}

	return current;
}

void loadcfg()
{
	struct strll *cfgchargeband=NULL;
	struct cfginf pppcfg[]={{"CHARGEBAND", 4, NULL},
				{"MINCHARGE", 3, NULL}};
	struct strll *current;
	int result;

	pppcfg[0].var=&cfgchargeband;
	pppcfg[1].var=&mincharge;

	result=readcfg(CFGFILE, pppcfg);

	current=cfgchargeband;
	while (current!=NULL) {
		chargelist=addchargeband(chargelist, current->str);
		current=current->next;
	}
}

char *timetohms(time_t timper, int sht)
{
	static char tmp[100];

	if (sht) {
		snprintf(tmp, 99, "%02ld:%02ld:%02ld", timper/3600, (timper%3600)/60, (timper%3600)%60);
	} else if (timper<60) {
		snprintf(tmp, 99, "%ld seconds", timper);
	} else if (timper<3600) {
		snprintf(tmp, 99, "%ld minutes and %ld seconds", timper/60, timper%60);
	} else {
		snprintf(tmp, 99, "%ld hours, %ld minutes and %ld seconds", timper/3600, (timper%3600)/60, (timper%3600)%60);
	}

	return tmp;
}

int charge(time_t starttime, time_t endtime)
{
	int tmpcharge;
	struct chargeband *curband;
	struct tm *timeinfo;

	timeinfo=localtime(&starttime);

	/* Make tmpcharge equal minutes online */
	tmpcharge = ((endtime-starttime)+59)/60;

	curband=chargelist;

	while (curband!=NULL &&
		!(((curband->startday==-1) || (timeinfo->tm_wday>=curband->startday
					 && timeinfo->tm_wday<=curband->endday)) &&
		((curband->starthour==-1) || (timeinfo->tm_hour>=curband->starthour
					 && timeinfo->tm_hour<=curband->endhour)) &&
		((curband->startmin==-1) || (timeinfo->tm_min>=curband->startmin
					 && timeinfo->tm_min<=curband->endmin)))
		) {
		curband=curband->next;
	}

	if (curband==NULL) {
		puts("Error: No charging band for time in charge()");
		exit(1);
	}

	tmpcharge *= curband->cost;

	/* Minimum call charge */
	if (tmpcharge<mincharge) tmpcharge=mincharge;

	return tmpcharge;
}

void calcharge(const char *name, int itemize)
{
	FILE *chgfile;
	struct charginginf chgdata;
	time_t totaltime;
	time_t lasttime;
	unsigned long totalcalls;
	unsigned long totalcharge,tmpcharge;
	int incharge;
	char *tmps,*tmpe;

	totaltime=lasttime=totalcalls=totalcharge=incharge=0;

	if ((chgfile=fopen(CHARGINGFILE,"r"))==NULL) {
		printf("calcharge(): Couldn't open %s\n", CHARGINGFILE);
		exit(1);
	}

	if (itemize) {
		if (html) {
			printf("<TABLE>\n");
			puts("<TR><TH>User</TH><TH>Start time</TH><TH>Length</TH><TH>Cost</TH></TR>");
		} else {
			puts("User      Start time                Length    Cost");
		}
	}

	fread(&chgdata, 1, sizeof(chgdata), chgfile);

	while (!feof(chgfile)) {
		if (name==NULL || strcmp(chgdata.name, name)==0) {
			if (chgdata.action==0) {
				lasttime=chgdata.time;
				++totalcalls;
				incharge=1;
			} else if (chgdata.action==1 && incharge) {
				tmpcharge=charge(lasttime, chgdata.time);

				if (itemize) {
					tmps=strdup(ctime(&lasttime));
					tmpe=strdup(ctime(&chgdata.time));
					tmps[strlen(tmps)-1]=0;
					tmpe[strlen(tmpe)-1]=0;
					if (html) {
						printf("<TR><TD>%s</TD><TD>%s</TD><TD>%s</TD><TD>%ldp</TD></TR>\n", chgdata.name, tmps, timetohms(chgdata.time-lasttime, 1), tmpcharge);
					} else {
						printf("%-8s  %s  %-8s  %ldp\n", chgdata.name, tmps, timetohms(chgdata.time-lasttime, 1), tmpcharge);
					}
					free(tmps);
					free(tmpe);
				}

				totalcharge+=tmpcharge;

				totaltime+=(chgdata.time-lasttime);
				incharge=0;
			}
		}

		fread(&chgdata, 1, sizeof(chgdata), chgfile);
	}

	fclose(chgfile);
	
	if (itemize) {
		if (html) {
			printf("</TABLE>\n\n");
		} else {
			putc('\n', stdout);
		}
	}

	if (name==NULL)
		printf("Global call details:%s\n", html ? "<BR>" : "");
	else
		printf("Call details for %s:%s\n", name, html ? "<BR>" : "");

	printf("\tTotal time online: %s%s\n", timetohms(totaltime,0), html ? "<BR>" : "");

	printf("\tTotal calls: %ld%s\n", totalcalls, html ? "<BR>" : "");
	printf("\tTotal charge: %ldp%s\n", totalcharge, html ? "<BR>" : "");
}

void curcharge(void)
{
	FILE *chgfile;
	struct charginginf chgdata;
	time_t totaltime;
	time_t lasttime;
	unsigned long tmpcharge;

	totaltime=lasttime=0;

	if ((chgfile=fopen(CHARGINGFILE,"r"))==NULL) {
		printf("curcharge(): Couldn't open %s\n", CHARGINGFILE);
		exit(1);
	}

	fseek(chgfile, -sizeof(chgdata), SEEK_END);
	fread(&chgdata, 1, sizeof(chgdata), chgfile);

	if (chgdata.action==0) {
		tmpcharge=charge(chgdata.time, time(NULL));

		printf("Current call has lasted %s, cost so far %ldp\n",
			timetohms(time(NULL)-chgdata.time, 0), tmpcharge);

	} else if (chgdata.action==1) {
		printf("Not currently in a call.\n");
	}

	fclose(chgfile);
}

void lastcharge(void)
{
	FILE *chgfile;
	struct charginginf start,end;
	time_t totaltime;
	time_t lasttime;
	unsigned long tmpcharge;

	totaltime=lasttime=0;

	if ((chgfile=fopen(CHARGINGFILE,"r"))==NULL) {
		printf("curcharge(): Couldn't open %s\n", CHARGINGFILE);
		exit(1);
	}

	fseek(chgfile, -(2*sizeof(struct charginginf)), SEEK_END);
	fread(&start, 1, sizeof(start), chgfile);
	fread(&end, 1, sizeof(end), chgfile);

	if (start.action==0 && end.action==1) {
		tmpcharge=charge(start.time,end.time);

		printf("Last call lasted %s and cost %ldp\n",
			timetohms(end.time-start.time, 0), tmpcharge);

	} else if (end.action==0) {
		printf("Currently in a call.\n");
	} else {
		printf("Unknown data.\n");
	}

	fclose(chgfile);
}

void fixcharging()
{
	FILE *oldchgfile, *newchgfile;
	struct charginginf chgdata, oldchgdata;
	int incharge;

	if ((oldchgfile=fopen(CHARGINGFILE,"r"))==NULL) {
		printf("fixcharging(): Couldn't open %s\n", CHARGINGFILE);
		exit(1);
	}

	if ((newchgfile=fopen(TMPCHARGINGFILE,"w"))==NULL) {
		printf("fixcharging(): Couldn't open %s\n", TMPCHARGINGFILE);
		exit(1);
	}

	fread(&chgdata, 1, sizeof(chgdata), oldchgfile);
	incharge=0;

	while (!feof(oldchgfile)) {
		if (chgdata.action==0 && !incharge)
			incharge=1;
		else if (chgdata.action==0 && incharge) {
			printf("Found new dialup action when previous one was never hung up. Fixing.\n");
			oldchgdata.time=chgdata.time-1;
			oldchgdata.action=1;
			fwrite(&oldchgdata, 1, sizeof(oldchgdata), newchgfile);
		} else if (chgdata.action==1) incharge=0;
	
		fwrite(&chgdata, 1, sizeof(chgdata), newchgfile);
		oldchgdata=chgdata;
		fread(&chgdata, 1, sizeof(chgdata), oldchgfile);
	}

	fclose(newchgfile);
	fclose(oldchgfile);

	rename(TMPCHARGINGFILE, CHARGINGFILE);
}

void listcharging(const char *name)
{
	FILE *chgfile;
	struct charginginf chgdata;

	if ((chgfile=fopen(CHARGINGFILE,"r"))==NULL) {
		printf("listcharging(): Couldn't open %s\n", CHARGINGFILE);
		exit(1);
	}

	if (html) {
		printf("<TABLE>\n");
		printf("<TR><TH>Name</TH><TH>Action</TH><TH>Time</TH></TR>\n");
	} else {
		printf("Name     Action  Time\n");
	}

	fread(&chgdata, 1, sizeof(chgdata), chgfile);

	while (!feof(chgfile)) {
		if (name==NULL || strcmp(name, chgdata.name)==0) {
			if (html) {
				printf("<TR><TD>%s</TD><TD>%s</TD><TD>%s</TD></TR>",chgdata.name, (chgdata.action==0) ?
				"started" : "hungup ", ctime(&chgdata.time));
			} else {
				printf("%-8s %s %s",chgdata.name, (chgdata.action==0) ?
				"started" : "hungup ", ctime(&chgdata.time));
			}
		}

		fread(&chgdata, 1, sizeof(chgdata), chgfile);
	}

	fclose(chgfile);
	
	if (html) printf("</TABLE>\n");
}

void startstop(const char *name, unsigned char action)
{
	FILE *chgfile;
	struct charginginf chgdata;

	if ((chgfile=fopen(CHARGINGFILE,"a"))==NULL) {
		printf("startstop(): Couldn't open %s\n", CHARGINGFILE);
		exit(1);
	}

	strncpy(chgdata.name, name, 8);
	chgdata.name[8]=0;
	chgdata.action=action;
	chgdata.time=time(NULL);

	fwrite(&chgdata, 1, sizeof(chgdata), chgfile);

	fclose(chgfile);
}

void usage()
{
	printf("pppcharge v%s by Jonathan McDowell\n", VERSION);
	printf("usage: pppcharge ACTION [--user <user>] [--html]\n\n");
	printf("  --start         Start charging <user> (from --user parameter)\n");
	printf("  --stop          Stop charging <user> (from --user parameter)\n");
	printf("  --bill          Calculate bill for given user, or total bill if no user\n");
	printf("                  given.\n");
	printf("  --current       Calculate the cost of the current call\n");
	printf("  --help          Display this usage information.\n");
	printf("  --itemize       Display an itemized bill for a given user, or all calls if\n");
	printf("                  no user given.\n");
	printf("  --last          Calculate the cost of the last call\n");
	printf("  --list          List only entries for given user, or all entries if no user\n");
	printf("                  given.\n\n");
	printf("  --user <user>   Specify which user to perform the actions above for.\n");
	printf("  --html          Use html formatting codes when outputting data. Note that\n");
	printf("                  this does not output header or footer html.\n");
}

int main(int argc, char **argv)
{
	int option;
	char *user=NULL;
	int function=0;
	
	loadcfg();

	while (1) {
		int longoption;
		static struct option long_options[]={
			{"bill", 0, NULL, 'b'},
			{"current", 0, NULL, 'c'},
			{"help", 0, NULL, 'h'},
			{"html", 0, &html, 1},
			{"itemize", 0, NULL, 'i'},
			{"last", 0, NULL, 'p'},
			{"list", 0, NULL, 'l'},
			{"start", 0, NULL, 'x'},
			{"stop", 0, NULL, 'y'},
			{"user", 1, NULL, 'u'},
			{"verbose", 0, &verbose, 1},
			{0,0,0,0}
		};

		option=getopt_long(argc, argv, "bcfhltu:", long_options, &longoption);

		if (option==-1) break;

		switch (option) {
		case 'b':	function+=1; /* bill */
				break;
		case 'c':	function+=2; /* current */
				break;
		case 'f':	function+=4; /* fix */
				break;
		case 'h':	function+=8; /* usage */
				break;
		case 'i':	function+=256; /* itemize */
				break;
		case 'l':	function+=16; /* list */
				break;
		case 'p':	function+=128; /* last */
				break;
		case 'u':	if (user==NULL)
					user=strdup(optarg);
				else
					printf("Already set user!\n");
				break;
		case 'x':	function+=32;
				break;
		case 'y':	function+=64;
				break;
		}
	}

#ifdef DEBUG
	if (html) {
		puts("<!-- Development version -->");
	} else {
		puts("***DEVELOPMENT VERSION***");
	}
#endif

	switch (function) {
	case 0:		puts("No function specified.");
			puts("Need one of: start, stop, bill, current, help, itemize, last or list");
			break;
	case 1:		calcharge(user, 0);
			break;
	case 2:		curcharge();
			break;
	case 4:		fixcharging();
			break;
	case 8:		usage();
			break;
	case 16:	listcharging(user);
			break;
	case 32:	if (user==NULL || strlen(user)==0) {
				puts("Must set user to start charging.");
				exit(1);
			}
			startstop(user, 0);	/* start */
			break;
	case 64:	if (user==NULL || strlen(user)==0) {
				puts("Must set user to stop charging.");
				exit(1);
			}
			startstop(user, 1);	/* stop */
			break;
	case 128:	lastcharge();
			break;
	case 256:	calcharge(user, 1);
			break;
	default:	printf("Can't specify more than one function!\n");
			break;
	}

	return 0;
}
